#include <stdlib.h>
#include <stdio.h>
#include <i86.h>
#include <conio.h>

int oplbase = 0x388;

void delayreg() {
    int timeout = 20; while (--timeout) inp(0x80);
}

void delaydata() {
    int timeout = 60; while (--timeout) inp(0x80);
}

int opl_in(int base, int reg) {
    outp(base + 0, reg);
    delayreg();
    int rtn = inp(base + 1);
    delaydata();
    return rtn;
}

void opl_out(int base, int reg, int data) {
    outp(base + 0, reg);
    delayreg();
    outp(base + 1, data);
    delaydata();
}

void opl_reset(int base) {
    for (int reg = 0; reg < 0x100; reg++) {
        opl_out(base, reg, 0);
    }
}

enum {
    // chip type
    CHIP_NONE,
    CHIP_OPL2_SINGLE,
    CHIP_OPL2_DUAL,
    CHIP_OPL3,

    // chip flags
    CHIP_RW_REGS          = (1 << 0),
    CHIP_OPL2_ALIASED     = (1 << 1),
    CHIP_LIKELY_CQM       = (1 << 2),
};

const char *chiptype_name[] = {
    "None", "OPL2", "OPL2 Dual", "OPL3"
};

int main(int argc, char *argv[]) {
    printf("------------------------\n");
    int chiptype = CHIP_NONE, chipflags = 0;

    if (argc >= 2) oplbase = strtol(argv[1], NULL, 16);

    // test if OPL is present
    opl_out(oplbase, 4, 0x60);      // reset T1/T2
    opl_out(oplbase, 4, 0x80);      // reset IRQ
    opl_out(oplbase+2, 4, 0x60);    // do the same with (possible) second OPL2
    opl_out(oplbase+2, 4, 0x80);
    int stat0 = inp(oplbase);
    opl_out(oplbase, 2, 0xFF);  // set T1
    opl_out(oplbase, 4, 0x21);  // enable T1
    int timeout = 1024; while (--timeout) inp(0x80); // delay
    int stat1 = inp(oplbase);
    // be nice and stop the timers
    opl_out(oplbase, 4, 0x60);
    opl_out(oplbase, 4, 0x80);

    printf("OPL detect status = %02X -> %02X\n", stat0, stat1);
    bool opl_detected = (((stat0 & 0xE0) == 0) && ((stat1 & 0xE0) == 0xC0));
    printf("OPL detect: %s\n", opl_detected ? "SUCCESS": "FAIL");
    if (opl_detected) chiptype = CHIP_OPL2_SINGLE;

    // chick if likely to be an OPL3
    if ((stat1 & 0x06) == 0) {
        printf("Chip is likely to be an OPL3\n");
        chiptype = CHIP_OPL3;
    }

    // check for address aliasing
    int stat38A = inp(oplbase+2);
    bool single_opl2_aliased = (stat38A == stat1);
    if (single_opl2_aliased && chiptype != CHIP_NONE) {
        printf("single OPL2 is decoded at both %02X and %02X\n", oplbase, oplbase+2);
        chipflags |= CHIP_OPL2_ALIASED;
    }

    if (!single_opl2_aliased) {
        // check for dual OPL2 config
        opl_out(oplbase+2, 2, 0xFF);  // set T1
        opl_out(oplbase+2, 4, 0x21);  // enable T1
        int timeout = 1024; while (--timeout) inp(0x80); // delay
        int stat2 = inp(oplbase+2);
        printf("2nd OPL2 detect status = %02X -> %02X\n", stat38A, stat2);
        bool opl_detected = (((stat38A & 0xE0) == 0) && ((stat2 & 0xE0) == 0xC0));
        if (opl_detected) {
            printf("2nd OPL2 detect: %s\n", opl_detected ? "SUCCESS": "FAIL");
            chiptype = CHIP_OPL2_DUAL;
        }
        // be nice and stop the timers
        opl_out(oplbase+2, 4, 0x60);
        opl_out(oplbase+2, 4, 0x80);
    }

    // print current detection status
    printf("Chip type: %s\n", chiptype_name[chiptype]);

    // return if nothing is found
    if (chiptype == CHIP_NONE) return 0;

    // "reset" chips
    opl_reset(oplbase);
    if (chiptype >= CHIP_OPL2_DUAL) opl_reset(oplbase+2);

    for (int bank = 0; bank < 2; bank++)
    for (int r = 0; r < 2; r++)
    for (int b = 0; b < 8; b++) {
        int base = oplbase + (bank<<1);
        printf("Test Bank%d REG%02X = %02X: ", bank, r, (1 << b));

        // check test register behavior (OPL3 only)
        opl_out(oplbase, 4, 0x60);      // reset T1/T2
        opl_out(oplbase, 4, 0x80);      // reset IRQ
        int tst0 = inp(oplbase);
        opl_out(base, 1, (1 << b));
        int tst1 = inp(oplbase);
        timeout = 1024; while (--timeout) inp(0x80); // delay
        int tst2 = inp(oplbase);
        opl_out(base, 1, 0);
        int tst3 = inp(oplbase);
        printf("%02X %02X %02X %02X ", tst0, tst1, tst2, tst3);

        // restart timer
        opl_out(oplbase, 2, 0xFF);  // set T1
        opl_out(oplbase, 4, 0x21);  // enable T1
        timeout = 1024; while (--timeout) if ((inp(oplbase) & 0x40)) break; // delay

        // and check test register again
        tst0 = inp(oplbase);
        opl_out(base, 1, (1 << b));
        tst1 = inp(oplbase);
        timeout = 1024; while (--timeout) inp(0x80); // delay
        tst2 = inp(oplbase);
        opl_out(base, 1, 0);
        tst3 = inp(oplbase);
        printf("%02X %02X %02X %02X ", tst0, tst1, tst2, tst3);

        // set timer 2
        opl_out(oplbase, 3, 0xFF);  // set T2
        opl_out(oplbase, 4, 0x42);  // enable T2
        timeout = 4096; while (--timeout) if ((inp(oplbase) & 0x20)) break; // delay

        // and check test register again
        tst0 = inp(oplbase);
        opl_out(base, 1, (1 << b));
        tst1 = inp(oplbase);
        timeout = 1024; while (--timeout) inp(0x80); // delay
        tst2 = inp(oplbase);
        opl_out(base, 1, 0);
        tst3 = inp(oplbase);
        printf("%02X %02X %02X %02X\n", tst0, tst1, tst2, tst3);
    }

    // cleanup
    opl_out(oplbase, 4, 0x60);
    opl_out(oplbase, 4, 0x80);
    opl_reset(oplbase);
    if (chiptype >= CHIP_OPL2_DUAL) opl_reset(oplbase+2);

    return 0;
}
