#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <i86.h>
#include <conio.h>

int oplbase = 0x388;

void delayreg() {
    int timeout = 20; while (--timeout) _asm in al, 0x80;
}

void delaydata() {
    int timeout = 60; while (--timeout) _asm in al, 0x80;
}

int opl_in(int base, int reg) {
    outp(base + 0, reg);
    delayreg();
    int rtn = inp(base + 1);
    delaydata();
    return rtn;
}

void opl_out(int base, int reg, int data) {
    outp(base + 0, reg);
    delayreg();
    outp(base + 1, data);
    delaydata();
}

void opl_reset(int base) {
    for (int reg = 0; reg < 0x100; reg++) {
        opl_out(base, reg, 0);
    }
}

enum {
    // chip type
    CHIP_NONE,
    CHIP_OPL2_SINGLE,
    CHIP_OPL2_DUAL,
    CHIP_OPL3,
    CHIP_OPL3L,
    CHIP_ESFM,
    CHIP_CQM
};

const char *chiptype_name[] = {
    "None",
    "OPL2 (YM3812)",
    "OPL2 (YM3812) Dual",
    "OPL3 (YMF262) or compatible",
    "OPL3-L (YMF289), OPL3-SA (YMF71x) or DS-1 (YMF7x4)",
    "ESFM",
    "Creative CQM"
};

int opl_detect(int oplbase) {
    int chiptype = CHIP_NONE;
    int timeout = 0;

    // test if OPL is present
    opl_out(oplbase, 4, 0x60);      // reset T1/T2
    opl_out(oplbase, 4, 0x80);      // reset IRQ
    opl_out(oplbase+2, 4, 0x60);    // do the same with (possible) second OPL2
    opl_out(oplbase+2, 4, 0x80);
    int stat0 = inp(oplbase);
    opl_out(oplbase, 2, 0xFF);  // set T1
    opl_out(oplbase, 4, 0x21);  // enable T1
    timeout = 1024; while (--timeout) _asm in al, 0x80; // delay
    int stat1 = inp(oplbase);
    // be nice and stop the timers
    opl_out(oplbase, 4, 0x60);
    opl_out(oplbase, 4, 0x80);

    // check for OPL2
    bool opl_detected = (((stat0 & 0xE0) == 0) && ((stat1 & 0xE0) == 0xC0));
    if (opl_detected) chiptype = CHIP_OPL2_SINGLE; else return chiptype;

    // reset registers (this will also reset the LSI test register used later)
    opl_reset(oplbase);

    // check if likely to be an OPL3
    if ((stat1 & 0x06) == 0) chiptype = CHIP_OPL3;

    // check for address aliasing
    int stat38A = inp(oplbase+2);

    // if not aliased, check for dual OPL2 config
    if (stat38A != stat1) {
        // setup timers
        opl_out(oplbase+2, 2, 0xFF);  // set T1
        opl_out(oplbase+2, 4, 0x21);  // enable T1
        timeout = 1024; while (--timeout) _asm in al, 0x80; // delay

        // read status
        int stat2 = inp(oplbase+2);
        bool opl_detected = (((stat38A & 0xE0) == 0) && ((stat2 & 0xE0) == 0xC0));
        if (opl_detected) chiptype = CHIP_OPL2_DUAL;

        // be nice and stop the timers
        opl_out(oplbase+2, 4, 0x60);
        opl_out(oplbase+2, 4, 0x80);
    }

    // if chip is OPL3-like, determine the exact type
    if (chiptype >= CHIP_OPL3) {
        // reset 2nd bank
        opl_reset(oplbase+2);

        // check if registers are R/W
        int rw0 = opl_in(oplbase, 2);
        opl_out(oplbase, 2, 0x55);
        int rw1 = opl_in(oplbase, 2);
        opl_out(oplbase, 2, 0xAA);
        int rw2 = opl_in(oplbase, 2);
        if ((rw0 == 0x0) && (rw1 == 0x55) && (rw2 == 0xAA)) {
            chiptype = CHIP_OPL3L;
        }
        opl_out(oplbase, 2, 0); // cleanup

        // try to enable ESFM native mode
        opl_out(oplbase+2, 0x05, 0x80);
        // try to read compatibility register now
        opl_out(oplbase+2, 0x05, 0x05);         // select ESFM reg 0x505
        if (inp(oplbase+1) == 0x80) {
            // it's a ESFM!
            chiptype = CHIP_ESFM;
            // return to OPL3-compatible mode
            outp(oplbase, 0);
        }

        // check for Creative CQM - test register bit 0 enables
        // unknown "counter readback" in status register
        if (chiptype == CHIP_OPL3) {            // CQM register array is write-only
            int tst0 = inp(oplbase) & 0x1F;
            opl_out(oplbase, 1, 1);             // set LSI test bit 0

            // check if status register bits 0-5 are changed within timeout
            timeout = 256; while (--timeout) {
                if ((inp(oplbase) & 0x1F) != tst0) break;
            }

            // if not timed out, we have a CQM
            if (timeout != 0) chiptype = CHIP_CQM;
            opl_out(oplbase, 1, 0);             // cleanup
        }
    }

    if (chiptype != CHIP_NONE) {
        // cleanup
        opl_out(oplbase, 4, 0x60);  // reset T1/T2
        opl_out(oplbase, 4, 0x80);  // reset IRQ
        opl_reset(oplbase);         // reset registers
        if (chiptype >= CHIP_OPL2_DUAL)
            opl_reset(oplbase+2);
    }

    return chiptype;
}

int main(int argc, char *argv[]) {
    int timeout = 0;
    printf("------------------------\n");

    if (argc >= 2) oplbase = strtol(argv[1], NULL, 16);

    // detect!
    int chiptype = opl_detect(oplbase);

    // print current detection status
    printf("Chip type: %s\n", chiptype_name[chiptype]);

    return 0;
}
